# -*- encoding: utf-8 -*-

from urllib import urlencode
import urllib
import urlparse
from types import FunctionType
import tornado.web


def decorate_all(decorator_list):
    def is_method_need_to_decorate(func_name, func_obj, check_param):
        """check if an object should be decorated"""
        methods = ["get", "head", "post", "put", "delete"]
        return (func_name in methods and
                isinstance(func_obj, FunctionType) and
                getattr(func_obj, check_param, True))

    """decorate all instance methods (unless excluded) with the same decorator"""
    class DecorateAll(type):
        def __new__(cls, name, bases, dct):
            for func_name, func_obj in dct.iteritems():
                for item in decorator_list:
                    decorator, check_param = item
                    if is_method_need_to_decorate(func_name, func_obj, check_param):
                        dct[func_name] = decorator(dct[func_name])
            return super(DecorateAll, cls).__new__(cls, name, bases, dct)

        def __setattr__(self, func_name, func_obj):
            for item in decorator_list:
                decorator, check_param = item
                if is_method_need_to_decorate(func_name, func_obj, check_param):
                    func_obj = decorator(func_obj)
            super(DecorateAll, self).__setattr__(func_name, func_obj)
    return DecorateAll


def make_list(val):
    if isinstance(val, list):
        return val
    else:
        return [val]


def real_ip(request):
    return (request.headers.get("X-Real-Ip", None) or request.headers.get("X-Forwarded-For", None)
            or request.remote_ip or '127.0.0.1')


HTTPError = tornado.web.HTTPError


def update_url(url, update_args=None, remove_args=None):
    scheme, sep, url_new = url.partition('://')
    if len(scheme) == len(url):
        scheme = ''
    else:
        url = '//' + url_new

    url_split = urlparse.urlsplit(url.encode('utf-8') if isinstance(url, unicode) else url)
    query_dict = urlparse.parse_qs(url_split.query, keep_blank_values=True)

    # add args
    if update_args:
        query_dict.update(update_args)
        # remove args
    if remove_args:
        query_dict = dict([(k, query_dict.get(k)) for k in query_dict if k not in remove_args])

    query = urlencode(query_dict, doseq=True)
    # specific case without net location. Thx to maizy for this fuckin' case
    if url_split.netloc:
        return urlparse.urlunsplit((scheme, url_split.netloc, url_split.path, query, url_split.fragment))
    else:
        return ''.join([
            scheme,
            '://' if scheme else '',
            url_split.path,
            '?' if query else '',
            query,
            '#' if url_split.fragment else '',
            url_split.fragment
        ])


def make_qs(query_args):
    def _encode(s):
        if isinstance(s, unicode):
            return s.encode('utf-8')
        else:
            return s

    kv_pairs = []
    for (key, val) in query_args.iteritems():
        if val is not None:
            if isinstance(val, list):
                for v in val:
                    kv_pairs.append((key, _encode(v)))
            else:
                kv_pairs.append((key, _encode(val)))

    qs = urllib.urlencode(kv_pairs)

    return qs



# -*- coding: utf-8 -*-

import os
import re


def __parse_version_from_changelog():
    try:
        deb_path = os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))), 'debian', 'changelog')
        with open(deb_path, 'r') as changelog:
            regmatch = re.match(r'python-tortik \((.*)\).*', changelog.readline())
            return regmatch.groups()[0]
    except (IOError, AttributeError):
        return 'unknown_version'

version = __parse_version_from_changelog()



# -*- coding: utf-8 -*-

import httplib
from lxml import etree
from tornado.escape import json_decode

from tortik.util import HTTPError


def parse_xml(response):
    if response.code == 599 or response.buffer is None:
        raise HTTPError(httplib.SERVICE_UNAVAILABLE, 'Response timeout or no body buffer')
    try:
        parser = etree.XMLParser(remove_blank_text=True)
        return etree.parse(response.buffer, parser=parser)
    except etree.ParseError:
        raise HTTPError(httplib.SERVICE_UNAVAILABLE, 'Unable to parse xml')


def parse_json(response):
    if response.code == 599 or response.body is None:
        raise HTTPError(httplib.SERVICE_UNAVAILABLE, 'Response timeout or no body')
    try:
        result = json_decode(response.body)
    except ValueError:
        raise HTTPError(httplib.SERVICE_UNAVAILABLE, 'Unable to parse json')
    return result



# -*- coding: utf-8 -*-

import os
import sys
import time
import urlparse
import traceback
import hashlib
import random
from itertools import count
from functools import partial
from copy import copy
import json

import lxml.etree as etree
import tornado.web
import tornado.httpclient
from tornado.options import options, define
from tornado.escape import to_unicode
import tornado.gen
from jinja2 import Environment, PackageLoader

from tortik.util import decorate_all, make_list, real_ip, make_qs
from tortik.util.dumper import dump
from tortik.logger import PageLogger
from tortik.util.async import AsyncGroup
from tortik.util.parse import parse_xml, parse_json


define('debug_password', default=None, type=str, help='Password for debug')
define('debug', default=True, type=bool, help='Debug mode')
define('tortik_max_clients', default=200, type=int, help='Max clients (requests) for http_client')
define('tortik_timeout_multiplier', default=1.0, type=float, help='Timeout multiplier (affects all requests)')

_DEBUG_ALL = "all"
_DEBUG_ONLY_ERRORS = "only_errors"
_DEBUG_NONE = "none"

stats = count()


def _gen_requestid():
    return hashlib.md5("{}{}{}".format(os.getpid(), stats.next(), random.random())).hexdigest()


class RequestHandler(tornado.web.RequestHandler):
    decorators = [
        (tornado.web.asynchronous, 'asynchronous'),  # should be the last
    ]
    __metaclass__ = decorate_all(decorators)

    def initialize(self, *args, **kwargs):
        debug_pass = options.debug_password
        debug_agrs = self.get_arguments('debug')
        debug_arg_set = (len(debug_agrs) > 0 and debug_pass is not None and
                         (debug_pass == '' or debug_pass == debug_agrs[-1]))

        if debug_arg_set:
            self.debug_type = _DEBUG_ALL
        elif options.debug:
            self.debug_type = _DEBUG_ONLY_ERRORS
        else:
            self.debug_type = _DEBUG_NONE

        if self.debug_type != _DEBUG_NONE and not hasattr(RequestHandler, 'debug_loader'):
            environment = Environment(autoescape=True,
                                      loader=PackageLoader('tortik', 'templates'),
                                      extensions=['jinja2.ext.autoescape'],
                                      auto_reload=options.debug)

            environment.filters['split'] = lambda x, y: x.split(y)
            RequestHandler.debug_loader = environment

        self.error_detected = False

        self.request_id = self.request.headers.get('X-Request-Id', _gen_requestid())

        self.log = PageLogger(self.request, self.request_id, (self.debug_type != _DEBUG_NONE),
                              handler_name=(type(self).__module__ + '.' + type(self).__name__))

        self.responses = {}
        self.http_client = self.initialize_http_client()

        self.preprocessors = copy(self.preprocessors) if hasattr(self, 'preprocessors') else []
        self.postprocessors = copy(self.postprocessors) if hasattr(self, 'postprocessors') else []

        self.log.info('Using http client: %s' % repr(self.http_client))

        self._extra_data = {}

    @tornado.gen.coroutine
    def prepare(self):
        if self.preprocessors:
            start_time = time.time()
            yield map(lambda x: tornado.gen.Task(x, self), self.preprocessors)
            self.log.debug("Preprocessors completed in %.2fms", (time.time() - start_time)*1000.)

    @staticmethod
    def get_global_http_client():
        if not hasattr(RequestHandler, '_http_client'):
            RequestHandler._http_client = tornado.httpclient.AsyncHTTPClient(
                max_clients=options.tortik_max_clients)

        return RequestHandler._http_client

    def initialize_http_client(self):
        return self.get_global_http_client()

    def add(self, name, data):
        self._extra_data[name] = data

    def get_data(self):
        return self._extra_data

    def compute_etag(self):
        return None

    def on_finish(self):
        self.log.complete_logging(self.get_status())

    def write_error(self, status_code, **kwargs):
        if self.debug_type in [_DEBUG_ALL, _DEBUG_ONLY_ERRORS]:
            if 'exc_info' in kwargs:
                type, value, tb = kwargs['exc_info']
                self.log.error("Uncaught exception %s\n%r", self._request_summary(),
                               self.request, exc_info=(type, value, tb))

            if self._finished:
                return

            self.set_status(status_code)
            self.log.complete_logging(status_code)
            self.finish_with_debug()

            return True
        else:
            super(RequestHandler, self).write_error(status_code, kwargs)

    def finish_with_debug(self):
        self.set_header('Content-Type', 'text/html; charset=utf-8')
        if self.debug_type == _DEBUG_ALL:
            self.set_status(200)

        self.finish(RequestHandler.debug_loader.get_template('debug.html').render(
            data=self.log.get_debug_info(),
            output_data=self.get_data(),
            size=sys.getsizeof,
            get_params=lambda x: urlparse.parse_qs(x, keep_blank_values=True),
            pretty_json=lambda x: json.dumps(x, sort_keys=True, indent=4, ensure_ascii=False),
            pretty_xml=lambda x: etree.tostring(x, pretty_print=True, encoding=unicode),
            to_unicode=to_unicode,
            dumper=dump,
            format_exception=lambda x: "".join(traceback.format_exception(*x))
        ))

    def complete(self, output_data=None):
        def finished_cb(handler, data):
            handler.log.complete_logging(handler.get_status())
            if handler.debug_type == _DEBUG_ALL:
                self.finish_with_debug()
                return

            self.finish(data)

        if self.postprocessors:
            last = len(self.postprocessors) - 1

            def add_cb(index):
                if index == last:
                    return finished_cb
                else:
                    def _cb(handler, data):
                        self.postprocessors[index + 1](handler, data, add_cb(index + 1))
                    return _cb

            self.postprocessors[0](self, output_data, add_cb(0))
        else:
            finished_cb(self, output_data)

    def fetch_requests(self, requests, callback=None, stage='page'):
        self.log.stage_started(stage)
        requests = make_list(requests)

        def _finish_cb():
            self.log.stage_complete(stage)
            if callback is not None:
                callback()

        ag = AsyncGroup(_finish_cb, self.log.debug, name=stage)

        def _on_fetch(response, name):
            content_type = response.headers.get('Content-Type', '').split(';')[0]
            response.data = None
            try:
                if 'xml' in content_type:
                    response.data = parse_xml(response)
                elif content_type == 'application/json':
                    response.data = parse_json(response)
            except:
                self.log.warning('Could not parse response with Content-Type header')

            if response.data is not None:
                self.add(name, response.data)

            self.responses[name] = response
            self.log.request_complete(response)

        for req in requests:
            if isinstance(req, (tuple, list)):
                assert len(req) in (2, 3)
                req = self.make_request(name=req[0], method='GET', full_url=req[1],
                                        data=req[2] if len(req) == 3 else '')
            self.log.request_started(req)
            self.http_client.fetch(req, ag.add(partial(_on_fetch, name=req.name)))

    def make_request(self, name, method='GET', full_url=None, url_prefix=None, path='', data='', headers=None,
                     connect_timeout=0.5, request_timeout=2, follow_redirects=True, **kwargs):

        if (full_url is None) == (url_prefix is None):
            raise TypeError('make_request required path/url_prefix arguments pair or full_url argument')
        if full_url is not None and path != '':
            raise TypeError("Can't combine full_url and path arguments")

        scheme = 'http'
        query = ''
        body = None

        if full_url is not None:
            parsed_full_url = urlparse.urlsplit(full_url)
            scheme = parsed_full_url.scheme
            url_prefix = parsed_full_url.netloc
            path = parsed_full_url.path
            query = parsed_full_url.query

        if method in ['GET', 'HEAD']:
            parsed_query = urlparse.parse_qs(query)
            parsed_query.update(data if isinstance(data, dict) else urlparse.parse_qs(data))
            query = make_qs(parsed_query)
        else:
            body = make_qs(data) if isinstance(data, dict) else data

        headers = {} if headers is None else headers

        headers.update({
            'X-Forwarded-For': real_ip(self.request),
            'X-Request-Id': self.request_id,
            'Content-Type': headers.get('Content-Type', 'application/x-www-form-urlencoded')
        })

        req = tornado.httpclient.HTTPRequest(
            url=urlparse.urlunsplit((scheme, url_prefix, path, query, '')),
            method=method,
            headers=headers,
            body=body,
            connect_timeout=connect_timeout*options.tortik_timeout_multiplier,
            request_timeout=request_timeout*options.tortik_timeout_multiplier,
            follow_redirects=follow_redirects,
            **kwargs
        )
        req.name = name
        return req

    def add_preprocessor(self, preprocessor):
        self.preprocessors.append(preprocessor)

    def add_postprocessor(self, postprocessor):
        self.postprocessors.append(postprocessor)



# -*- coding: utf-8 -*-
from tornado.httputil import HTTPHeaders


def dump(obj):
    refs = set()

    def primitive(value, ptype='string'):
        return dict(type=ptype, value=value)

    def _make_dump(item):
        if isinstance(item, list):
            if len(item) == 0:
                return dict(type='array', value=[])
            if id(item) in refs:
                return primitive('<circular reference>')

            refs.add(id(item))
            return dict(type='array', value=map(lambda x: _make_dump(x), item))
        elif isinstance(item, dict):
            if not item:
                return dict(type='dict', value=dict())
            if id(item) in refs:
                return primitive('<circular reference>')
            refs.add(id(item))

            return dict(type='dict', value=dict((x, _make_dump(y)) for x, y in item.items()))
        elif isinstance(item, bool):
            return primitive('true' if item else 'false', 'bool')
        elif isinstance(item, basestring):
            return primitive(item)
        elif isinstance(item, (int, float)):
            return primitive(item, 'number')
        elif item is None:
            return primitive('null', 'null')
        else:
            return primitive(repr(item), 'string')

    return _make_dump(obj)


def request_to_curl_string(request):
    def _escape_apos(string):
        return string.replace("'", "'\"'\"'")

    try:
        if request.body:
            request.body.decode('ascii')
        is_binary_data = False
    except UnicodeError:
        is_binary_data = True

    curl_headers = HTTPHeaders(request.headers)
    if request.body and 'Content-Length' not in curl_headers:
        curl_headers['Content-Length'] = len(request.body)

    if is_binary_data:
        curl_echo_data = "echo -e {0} |".format(repr(request.body))
        curl_data_string = '--data-binary @-'
    else:
        curl_echo_data = ''
        curl_data_string = "--data '{0}'".format(_escape_apos(request.body)) if request.body else ''

    return "{echo} curl -X {method} '{url}' {headers} {data}".format(
        echo=curl_echo_data,
        method=request.method,
        url=request.url,
        headers=' '.join("-H '{0}: {1}'".format(k, _escape_apos(str(v))) for k, v in curl_headers.iteritems()),
        data=curl_data_string
    ).strip()



# -*- coding: utf-8 -*-
import time
import logging
import logging.handlers
from collections import OrderedDict
from datetime import datetime

from tornado.options import options, define

from tortik.util.dumper import request_to_curl_string

LOGGER_NAME = 'tortik'
_SKIP_EVENT = "skip_event"

tortik_log = logging.getLogger(LOGGER_NAME)


define('tortik_logformat', default='[%(process)s %(asctime)s %(levelname)s %(name)s] %(message)s')


class RequestIdFilter(logging.Filter):
    def filter(self, record):
        record.name = '.'.join(filter(None, [record.name, getattr(record, 'request_id', None)]))
        event_queue = getattr(record, 'event_queue', None)
        if event_queue is not None and getattr(record, _SKIP_EVENT, None) is None:
            created = record.created or time.time()
            event_queue[record] = {
                'created': str(datetime.fromtimestamp(created).time()),
                'msg': record.getMessage(),
                'exc_info': record.exc_info,
                'type': record.levelname}
        return True


class PageLogger(logging.LoggerAdapter):
    class StageInfo(object):
        def __init__(self, start):
            self.start = start
            self.end = None
            self.counter = 1

        def __str__(self):
            if self.end:
                return "{0:.2f}".format(1000 * (self.end - self.start))
            else:
                return "infinite"

    def __init__(self, request, request_id, use_debug, handler_name):
        self.debug_info = OrderedDict() if use_debug else None
        self.request = request
        logging.LoggerAdapter.__init__(
            self,
            tortik_log,
            {'request_id': str(request_id),
             'event_queue': self.debug_info})
        self.started = time.time()
        self.request_id = request_id
        self.handler_name = handler_name
        self.stages = OrderedDict()
        self.debug('Started {0} {1}'.format(request.method, request.uri), extra={_SKIP_EVENT: True})
        self._completed = False

    def process(self, msg, kwargs):
        if "extra" not in kwargs:
            kwargs["extra"] = self.extra
        else:
            kwargs["extra"].update(self.extra)
        return msg, kwargs

    # def cache_access_started(self, cache_name):
    #     self.debug("Requesting cache {0}".format(cache_name), extra={_SKIP_EVENT: True})
    #     if self.debug_info is not None:
    #         self.debug_info[cache_name] = {
    #             'type': 'cache',
    #             'started': time.time(),
    #             'name': cache_name
    #         }
    #
    # def cache_access_complete(self, name, value):
    #     if self.debug_info is not None:
    #         event = self.debug_info.get(name)
    #         if event is None:
    #             self.error("Cache access response is not found: {0}".format(name))
    #         else:
    #             now = time.time()
    #             event.update({
    #                 'completed': now,
    #                 'took': now - event['started'],
    #                 'value': repr(value),
    #                 })

    def request_started(self, request):
        self.debug("Requesting {0} {1}".format(request.method, request.url), extra={_SKIP_EVENT: True})
        if self.debug_info is not None:
            event = {
                'type': 'NET',
                'started': time.time(),
                'request': {
                    'url': request.url,
                    'method': request.method,
                    'headers': request.headers,
                    'data': request.body,
                    'curl': request_to_curl_string(request)
                }
            }
            self.debug_info[request] = event

    def request_complete(self, resp):
        if self.debug_info is not None:
            event = self.debug_info.get(resp.request, None)
            if event is None:
                self.error("Request for response is not found: {0}".format(resp.request.url))
            else:
                now = time.time()
                event.update({
                    'completed': now,
                    'took': now - event['started'],
                    'time_info': getattr(resp, 'time_info', None),
                    'response': {
                        'code': resp.code,
                        'headers': resp.headers,
                        'body': resp.body,
                        'data': resp.data
                    },
                })

        level = logging.DEBUG
        if 400 <= resp.code < 500:
            level = logging.INFO
        elif resp.code >= 500:
            level = logging.ERROR

        self.log(level, "Complete {0} {1} {2} in {3}ms".format(resp.code, resp.request.method, resp.request.url,
                                                               int(resp.request_time * 1000.0)),
                 extra={_SKIP_EVENT: True})

    def stage_started(self, stage_name):
        if stage_name not in self.stages:
            self.stages[stage_name] = PageLogger.StageInfo(time.time())
        else:
            self.warning("Stage {0} already started".format(stage_name))
            self.stages[stage_name].counter += 1

    def stage_complete(self, stage_name):
        if stage_name in self.stages:
            stage = self.stages[stage_name]
            stage.counter -= 1
            if stage.counter < 0:
                self.error("Freed stage {0} more times then it is needed".format(stage_name))
            elif stage.counter == 0:
                self.stages[stage_name].end = time.time()
        else:
            self.error("Stage {0} wasn't stared".format(stage_name))

    def complete_logging(self, status_code, additional_data=None):
        if self._completed:
            return

        if additional_data is None:
            additional_data = []

        data = [
            ('handler', self.handler_name),
            ('method', self.request.method),
            ('code', status_code),
            ('total', int(1000 * self.request.request_time()))  # <current_time> - <start> if request not finished yet
        ] + self.stages.items() + additional_data

        self.info('MONIK {0}'.format(' '.join('{0}={1}'.format(k, v) for (k, v) in data)))

        self._completed = True

    def get_debug_info(self):
        return self.debug_info.values()


def configure(logfile=None):
    if logfile:
        handler = logging.handlers.WatchedFileHandler(logfile)
        handler.setFormatter(logging.Formatter(options.tortik_logformat))
        tortik_log.addHandler(handler)

    tortik_log.addFilter(RequestIdFilter())
    tortik_log.setLevel(logging.DEBUG)



# -*- coding: utf-8 -*-

import time
from tornado.web import HTTPError
from tortik.logger import tortik_log


class AsyncGroup(object):
    """
    Grouping of several async requests and final callback in such way, that final callback is invoked after the last
     request is finished.
    """

    def __init__(self, finish_cb, log=None, name=None):
        self.counter = 0
        self.finished = False
        self.finish_cb = finish_cb
        self.log_fun = log if log is not None else tortik_log.debug
        self.name = name

        self.start_time = time.time()
        self.finish_time = None

        if self.name is not None:
            self.log_name = '{0} group'.format(self.name)
        else:
            self.log_name = 'group'

    def _log(self, msg, *args, **kw):
        self.log_fun(self.log_name + ": " + msg, *args, **kw)

    def finish(self):
        if not self.finished:
            self.finish_time = time.time()
            self._log('done in %.2fms', (self.finish_time - self.start_time)*1000.)
            self.finished = True

            try:
                self.finish_cb()
            finally:
                # prevent possible cycle references
                self.finish_cb = None

    def try_finish(self):
        if self.counter == 0:
            self.finish()

    def _inc(self):
        assert(not self.finished)
        self.counter += 1

    def _dec(self):
        self.counter -= 1
        self._log('%s requests pending', self.counter)

    def add(self, intermediate_cb):
        self._inc()

        def wrapper(*args, **kwargs):
            if not self.finished:
                try:
                    self._dec()
                    intermediate_cb(*args, **kwargs)
                except HTTPError:
                    raise
                except Exception:
                    self.try_finish()
                    raise
                else:
                    self.try_finish()
            else:
                self._log("Ignoring response because of already finished group")

        return wrapper

    def add_empty_cb(self):
        self._inc()

        def wrapper(*args, **kwargs):
            self._dec()
            self.try_finish()

        return wrapper



#!/usr/bin/env python
import os
from setuptools import setup, find_packages, Command
from setuptools.command.build_py import build_py

from tortik import version


def find_package_data(package, *directories):
    data_files = []
    current_dir = os.path.abspath(os.curdir)
    package_dir = os.path.dirname(__import__(package, fromlist=[""]).__file__)
    os.chdir(package_dir)

    for directory in directories:
        for dirpath, dirnames, filenames in os.walk(directory):
            [data_files.append(os.path.join(dirpath, f)) for f in filenames]

    os.chdir(current_dir)

    return data_files


class TestCommand(Command):
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        import nose
        nose.main(argv=['nosetests', 'tortik_tests/'])


class BuildHook(build_py):
    def run(self):
        build_py.run(self)

        build_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), self.build_lib, 'tortik')
        with open(os.path.join(build_dir, 'version.py'), 'w') as version_file:
            version_file.write('version = "{0}"\n'.format(version))

setup(
    name="tortik",
    version=version,
    description="Tortik - python tornado framework",
    long_description=open("README.rst").read(),
    url="https://github.com/glibin/tortik",
    download_url='https://github.com/glibin/tortik/tarball/{}'.format(version),
    packages=find_packages(exclude=['tortik_tests', 'tortik_tests.*']),
    package_data={
        "tortik": find_package_data("tortik", "templates"),
    },
    cmdclass={'test': TestCommand, 'build_py': BuildHook},
    requires=['tornado', 'lxml', 'nose', 'jinja2'],
)
